#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
File containing definitions for the notification system of the Workspaces REST API
"""
# pylint: disable=C0103, E0401, E0402

from http import HTTPStatus
from typing import Any, Dict, List

from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config, view_defaults

from workspaces.notification.schema import NotificationTemplate


@view_defaults(route_name="notifications", renderer="json")
class NotificationListRestService:
    """Handles requests involving notification templates"""

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def list_notification_templates(self) -> List[NotificationTemplate]:
        """
        Pyramid view to list all available notification templates
        :param request: GET request
        :return:
        """
        return self.request.notification_info.all_templates()


@view_defaults(route_name="notification", renderer="json")
class NotificationRestService:
    """Notification REST API"""

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="DELETE")
    def delete_notification_template(self) -> Response:
        """
        Pyramid view to remove a Notification template
        URL: /notify/{template_name}/delete

        :param request: DELETE request
        :return:
        """
        template_name = self.request.matchdict["template_name"]

        try:
            deleted = self.request.notification_info.delete_template(template_name)
        except Exception as e:
            return Response(
                "An error occurred deleting template: " + template_name + ".  \nError: " + str(e),
                status=HTTPStatus.INTERNAL_SERVER_ERROR,
            )

        return Response(deleted, status=HTTPStatus.OK)

    @view_config(route_name="send_notification", renderer="json")
    def send_notification(self) -> Dict[str, Any]:
        """
        Pyramid view to send a notification email
        :param request: POST request, expecting JSON parameters ["destination_email", "subject", "message"]
        :return:
        """

        json_packet = self.request.json_body
        template_name = self.request.matchdict["template_name"]

        r = self.request.notification_service.send_email(template_name, json_packet)

        return dict(message=f"Email sent{' with errors' if r != {} else ''}", errors=r)

    @view_config(request_method="POST")
    def update_notification_template(self) -> Response:
        """
        Change the content of a notification template.

        :return:
        """
        template_name = self.request.json_body["template_name"]
        template_text = self.request.json_body["template_text"]
        try:
            self.request.notification_info.update_template(template_name=template_name, template_text=template_text)
        except Exception as exc:
            return Response(
                "An error occurred updating template: " + template_name + ".  \nError: " + str(exc),
                status=HTTPStatus.INTERNAL_SERVER_ERROR,
            )

        return Response(json_body={"resp": f"template '{template_name}' has been updated."}, status=HTTPStatus.OK)

    @view_config(request_method="PUT")
    def create_notification_template(self) -> NotificationTemplate:
        return self.request.notification_info.create_template(
            self.request.json_body["template_name"],
            self.request.json_body["description"],
            self.request.json_body["content"],
        )

    @view_config(route_name="healthcheck", renderer="json")
    def get_healthcheck(self) -> Response:
        """
        Route for capability container to check notification is ready
        :param request: GET request
        :return:
        """
        return Response(
            status=HTTPStatus.OK, json_body={"healthcheck": f"Notification service returned {HTTPStatus.OK}"}
        )
